package Apache::Authn::RedmineRRD;

=head1 Apache::Authn::RedmineRRD

RedmineRRD - a mod_perl module to authenticate repository users directly
against the redmine core using a socket connection to the 
Redmine Repository Daemon (RRD)

Module based on original "Redmine.pm"

=head1 SYNOPSIS

This module allows anonymous users to browse public projects and
registred users to browse and commit their project. Authentication is
done directly against redmine, which includes LDAP and other 
authentication methods redmine supports.

=head1 INSTALLATION

For this to automagically work, you need to have a recent reposman.rb
(after r860) and if you already use reposman, read the last section to
migrate.

No additional perl modules are needed for this module. Just ensure that
you start the RRD Daemon on startup from the redmine root directory:

Read the comments in the serverscript "rrd" for further instructions.

This perl module needs at least RRD Version 0.82

=head1 CONFIGURATION

   ## This module has to be in your perl path
   ## eg:  /usr/lib/perl5/Apache/Authn/RedmineRRD.pm
   PerlLoadModule Apache::Authn::RedmineRRD
   <Location /svn>
     DAV svn
     SVNParentPath "/var/svn"

     AuthType Basic
     AuthName redmine
     Require valid-user

     PerlAccessHandler Apache::Authn::RedmineRRD::access_handler
     PerlAuthenHandler Apache::Authn::RedmineRRD::authen_handler

     ## access to redmine via RRD
     Redmine_RRD_SOCKET "/tmp/redmine_repository_daemon"
  </Location>

To be able to browse repository inside redmine, you must add something
like that :

   <Location /svn-private>
     DAV svn
     SVNParentPath "/var/svn"
     Order deny,allow
     Deny from all
     # only allow reading orders
     <Limit GET PROPFIND OPTIONS REPORT>
       Allow from redmine.server.ip
     </Limit>
   </Location>

and you will have to use this reposman.rb command line to create repository :

  reposman.rb --redmine my.redmine.server --svn-dir /var/svn --owner www-data -u http://svn.server/svn-private/

=head1 REPOSITORIES NAMING

A projet repository must be named with the projet identifier. In case
of multiple repositories for the same project, use the project identifier
and the repository identifier separated with a dot:

  /var/svn/foo
  /var/svn/foo.otherrepo

=head1 MIGRATION FROM OLDER RELEASES

If you use an older reposman.rb (r860 or before), you need to change
rights on repositories to allow the apache user to read and write
S<them :>

  sudo chown -R www-data /var/svn/*
  sudo chmod -R u+w /var/svn/*

And you need to upgrade at least reposman.rb (after r860).

=head1 GIT SMART HTTP SUPPORT

Git's smart HTTP protocol (available since Git 1.7.0) will not work with the
above settings. RedmineRRD.pm normally does access control depending on the HTTP
method used: read-only methods are OK for everyone in public projects and
members with read rights in private projects. The rest require membership with
commit rights in the project.

However, this scheme doesn't work for Git's smart HTTP protocol, as it will use
POST even for a simple clone. Instead, read-only requests must be detected using
the full URL (including the query string): anything that doesn't belong to the
git-receive-pack service is read-only.

To activate this mode of operation, add this line inside your <Location /git>
block:

  RedmineGitSmartHttp yes

Here's a sample Apache configuration which integrates git-http-backend with
a MySQL database and this new option:

   SetEnv GIT_PROJECT_ROOT /var/www/git/
   SetEnv GIT_HTTP_EXPORT_ALL
   ScriptAlias /git/ /usr/libexec/git-core/git-http-backend/
   <Location /git>
       Order allow,deny
       Allow from all

       AuthType Basic
       AuthName Git
       Require valid-user

       PerlAccessHandler Apache::Authn::RedmineRRD::access_handler
       PerlAuthenHandler Apache::Authn::RedmineRRD::authen_handler

       ## access to redmine via RRD
       Redmine_RRD_SOCKET "/tmp/redmine_repository_daemon"
       RedmineGitSmartHttp yes
    </Location>

Make sure that all the names of the repositories under /var/www/git/ have a
matching identifier for some project: /var/www/git/myproject and
/var/www/git/myproject.git will work. You can put both bare and non-bare
repositories in /var/www/git, though bare repositories are strongly
recommended. You should create them with the rights of the user running Redmine,
like this:

  cd /var/www/git
  sudo -u user-running-redmine mkdir myproject
  cd myproject
  sudo -u user-running-redmine git init --bare

Once you have activated this option, you have three options when cloning a
repository:

- Cloning using "http://user@host/git/repo(.git)" works, but will ask for the password
  all the time.

- Cloning with "http://user:pass@host/git/repo(.git)" does not have this problem, but
  this could reveal accidentally your password to the console in some versions
  of Git, and you would have to ensure that .git/config is not readable except
  by the owner for each of your projects.

- Use "http://host/git/repo(.git)", and store your credentials in the ~/.netrc
  file. This is the recommended solution, as you only have one file to protect
  and passwords will not be leaked accidentally to the console.

  IMPORTANT NOTE: It is *very important* that the file cannot be read by other
  users, as it will contain your password in cleartext. To create the file, you
  can use the following commands, replacing yourhost, youruser and yourpassword
  with the right values:

    touch ~/.netrc
    chmod 600 ~/.netrc
    echo -e "machine yourhost\nlogin youruser\npassword yourpassword" > ~/.netrc

=cut

use strict;
use warnings FATAL => 'all', NONFATAL => 'redefine';
use IO::Socket::UNIX qw( SOCK_STREAM );

use Apache2::Module;
use Apache2::Access;
use Apache2::ServerRec qw();
use Apache2::RequestRec qw();
use Apache2::RequestUtil qw();
use Apache2::Const qw(:common :override :cmd_how);
use APR::Pool ();
use APR::Table ();

my @directives = (
  {
    name => 'Redmine_RRD_SOCKET',
    req_override => OR_AUTHCFG,
    args_how => TAKE1,
    errmsg => 'Filename of socket file from RedmineRepositoryDaemon - usually /tmp/redmine_repository_daemon',
  },
  {
    name => 'RedmineGitSmartHttp',
    req_override => OR_AUTHCFG,
    args_how => TAKE1,
  },
);

sub Redmine_RRD_SOCKET { set_val('Redmine_RRD_SOCKET', @_); }

sub RedmineGitSmartHttp {
  my ($self, $parms, $arg) = @_;
  $arg = lc $arg;

  if ($arg eq "yes" || $arg eq "true") {
    $self->{RedmineGitSmartHttp} = 1;
  } else {
    $self->{RedmineGitSmartHttp} = 0;
  }
}

sub trim {
  my $string = shift;
  $string =~ s/\s{2,}/ /g;
  return $string;
}

sub set_val {
  my ($key, $self, $parms, $arg) = @_;
  $self->{$key} = $arg;
}

Apache2::Module::add(__PACKAGE__, \@directives);


my %read_only_methods = map { $_ => 1 } qw/GET HEAD PROPFIND REPORT OPTIONS/;

sub request_is_read_only {
  my ($r) = @_;
  my $cfg = Apache2::Module::get_config(__PACKAGE__, $r->server, $r->per_dir_config);

  # Do we use Git's smart HTTP protocol, or not?
  if (defined $cfg->{RedmineGitSmartHttp} and $cfg->{RedmineGitSmartHttp}) {
    my $uri = $r->unparsed_uri;
    my $location = $r->location;
    my $is_read_only = $uri !~ m{^$location/*[^/]+/+(info/refs\?service=)?git\-receive\-pack$}o;
    return $is_read_only;
  } else {
    # Standard behaviour: check the HTTP method
    my $method = $r->method;
    return defined $read_only_methods{$method};
  }
}

sub access_handler {
  my $r = shift;

  unless ($r->some_auth_required) {
      $r->log_reason("No authentication has been configured");
      return FORBIDDEN;
  }

  return OK unless request_is_read_only($r);

  my $project_id = get_project_identifier($r);

  $r->set_handlers(PerlAuthenHandler => [\&OK])
      if is_project_publicly_accessible($project_id, $r);

  return OK
}

sub authen_handler {
  my $r = shift;

  my ($res, $redmine_pass) =  $r->get_basic_auth_pw();
  return $res unless $res == OK;

  if (is_useraccess_allowed($r->user, $redmine_pass, $r)) {
      return OK;
  } else {
      $r->note_auth_failure();
      return DECLINED;
  }
}

# check if authentication is forced
sub is_authentication_forced {
  my $r = shift;


  #ok lets check if auth is forced:
  my $resp = ask_RRD("faut\n", $r);

  $resp =~ s/^2[0-9][0-9]\s-\s//; # cleanup result
  my $ret = ($resp eq "True");

  $ret;
}

sub is_project_publicly_accessible {
  my $project_id = shift;
  my $r = shift;

  my $access_mode = request_is_read_only($r) ? "Read_access" : "Write_access";

  if (is_authentication_forced($r)) {
      return 0;
    }
  my $resp = ask_RRD("prac ".$project_id."\n", $r);

  #ok lets check if access is permitted:
  #(read and write response are permitted everything else is invalid)
  $resp =~ s/^2[0-9][0-9]\s-\s//; # cleanup result
  my $ro = $resp;
  my $wo = $resp;
  $ro =~ s/[,\s]//g;
  $wo =~ s/[,\s]//g;
  $ro =~ s/Write_access//;
  $wo =~ s/Read_access//;
  my $ret = ($ro eq $access_mode) || ($wo eq "Write_access"); #read access fix (write permission includes reading)
  $ret;
}

sub is_useraccess_allowed {
  my $redmine_user = shift;
  my $redmine_pass = shift;
  my $r = shift;

  my $project_id  = get_project_identifier($r);

  my $access_mode = request_is_read_only($r) ? "Read_access" : "Write_access";

  my $cfg = Apache2::Module::get_config(__PACKAGE__, $r->server, $r->per_dir_config);

  my $resp = ask_RRD("auth ".$project_id.",".$redmine_user.",".$redmine_pass."\n", $r);

  #ok lets check if access is permitted:
  #(read and write response are permitted everything else is invalid)
  $resp =~ s/^2[0-9][0-9]\s-\s//; # cleanup result
  my $ro = $resp;
  my $wo = $resp;
  $ro =~ s/[,\s]//g;
  $wo =~ s/[,\s]//g;
  $ro =~ s/Write_access//;
  $wo =~ s/Read_access//;
#  $r->log_reason("RRD auth_resp: AM:".$access_mode." RO:".$ro." WO:".$wo);
  my $ret = ($ro eq $access_mode) || ($wo eq "Write_access"); #read access fix (write permission includes reading)
  $ret;
}

sub get_project_identifier {
    my $r = shift;

    my $cfg = Apache2::Module::get_config(__PACKAGE__, $r->server, $r->per_dir_config);
    my $location = $r->location;
    $location =~ s/\.git$// if (defined $cfg->{RedmineGitSmartHttp} and $cfg->{RedmineGitSmartHttp});
    my ($identifier) = $r->uri =~ m{$location/*([^/.]+)};
    $identifier;
}

sub ask_RRD {
  my $req = shift;
  my $r = shift;

  my $cfg = Apache2::Module::get_config(__PACKAGE__, $r->server, $r->per_dir_config);

  my $resp; # response - answer from RRD server
  my $ro;
  my $wo; 
  my $socket_path = $cfg->{Redmine_RRD_SOCKET};
#  $r->log_reason("RRD socket path: ".$socket_path);

  my $socket = IO::Socket::UNIX->new(
     Type => SOCK_STREAM,
     Peer => $socket_path,
  )
     or die("Can't connect to RRD server: $! Please start your RRD server daemon or correct the socket path\n");

  $socket->send("show ver\n");
  $socket->flush();
  chomp( $resp = <$socket> );
  $resp =~ s/^2[0-9][0-9]\s-\sVersion\s//; # cleanup result
  if ($resp >= 0.82) {# check -> ok
    # send authentification request to RRD: 
    $socket->send($req);
    $socket->flush();
    chomp( $resp = <$socket> );
  } else { # not acceptable -> invalid version 
    die("Invalid RRD Version - Please update/check this perl module!");
    $resp = "";
  }

  $socket->shutdown(2);
  $socket->close;
  $resp;
}

1;
